import requests

from requests.auth import HTTPBasicAuth

from python_mbills.base import MBillsBase
from python_mbills import constants
from python_mbills.exceptions import SignatureValidationException, TransactionDoesNotExist


class MBillsAPI(object):
    def __init__(self,
                 api_key,
                 shared_secret,
                 mbills_rsa_pub_key=None,
                 nonce_length=15,
                 api_endpoint="https://demo3.halcom.com/MBillsWS",
                 currency='EUR'):
        """
        Initialize the MBills API and Base class.
        :param api_key: 
        :param shared_secret: 
        :param mbills_rsa_pub_key: If None, no verification will be made 
        :param nonce_length: (default set to 15)
        :param api_endpoint: (default set to Apiary mock api)
        """
        self.api_endpoint = api_endpoint
        self.currency = currency
        self.base = MBillsBase(api_key, shared_secret, mbills_rsa_pub_key, nonce_length)

    def test_api_parameters_and_signature_verification(self):
        """
        Test if your configuration works correctly. 
        :return: 
        """
        url = "%s%s" % (self.api_endpoint, constants.TEST_ENDPOINT)

        username = self.base.get_username()
        password = self.base.get_password(username=username, request_url=url)

        response = requests.get(url, auth=HTTPBasicAuth(username=username, password=password))

        if not self.base.verify_response(response.json()):
            raise SignatureValidationException('Server signature verification has failed')

        return True

    def test_webhook(self):
        url = "%s%s" % (self.api_endpoint, constants.TEST_WEBHOOK)

        username = self.base.get_username()
        password = self.base.get_password(username=username, request_url=url)

        response = requests.get(url, auth=HTTPBasicAuth(username=username, password=password))

        if not self.base.verify_response(response.json()):
            raise SignatureValidationException('Server signature verification has failed')

        return True

    def create_new_sale(self, amount, purpose, payment_reference=None, order_id=None, channel_id=None, capture=True):
        """
        Create new sale.
        :param amount: 
        :param purpose: 
        :param payment_reference: 
        :param order_id: 
        :param channel_id: 
        :param capture: 
        :return: tuple (transaction_id, payment_token_number, status)
        """
        request_data = {
            "amount": self.base.convert_decimal_to_hundreds(amount),
            "currency": self.currency,
            "purpose": purpose,
            "capture": capture
        }

        if payment_reference:
            request_data['paymentreference'] = payment_reference

        if order_id:
            request_data['orderid'] = order_id

        if channel_id:
            request_data['channelid'] = channel_id

        url = "%s%s" % (self.api_endpoint, constants.NEW_SALE_ENDPOINT)

        username = self.base.get_username()
        password = self.base.get_password(username=username, request_url=url)

        response = requests.post(url, json=request_data, auth=HTTPBasicAuth(username=username, password=password))

        if not self.base.verify_response(response.json()):
            raise SignatureValidationException('Server signature verification has failed')

        response_json = response.json()

        return response_json.get('transactionid'), response_json.get('paymenttokennumber'), response_json.get('status')

    def fetch_transaction_status(self, transaction_id):
        """
        Get the transaction current status. 

        :param transaction_id: 
        :return: 
        """
        url = "%s%s%s/status" % (self.api_endpoint, constants.TRANSACTION_STATUS_ENDPOINT, transaction_id)

        username = self.base.get_username()
        password = self.base.get_password(username=username, request_url=url)

        response = requests.get(url, auth=HTTPBasicAuth(username=username, password=password))

        if response.status_code == 404:
            raise TransactionDoesNotExist('Wrong transaction ID!')

        if not self.base.verify_response(response.json()):
            raise SignatureValidationException('Server signature verification has failed')

        return response.json()

    def capture_sale(self, transaction_id, capture_amount, message=None):
        """
        Capture existing preauth.

        :param transaction_id:
        :param capture_amount:
        :param message:
        :return: status code
        """
        request_data = {
            "amount": self.base.convert_decimal_to_hundreds(capture_amount),
            "currency": self.currency,
            "message": message
        }

        url = "%s%s%s/capture" % (self.api_endpoint, constants.TRANSACTION_STATUS_ENDPOINT, transaction_id)

        username = self.base.get_username()
        password = self.base.get_password(username=username, request_url=url)

        response = requests.put(url, json=request_data, auth=HTTPBasicAuth(username=username, password=password))

        if response.status_code == 404:
            raise TransactionDoesNotExist('Wrong transaction ID!')

        if not self.base.verify_response(response.json()):
            raise SignatureValidationException('Server signature verification has failed')

        response_json = response.json()

        return response_json.get('status')

    def void_preauthorization(self, transaction_id, message=None):
        """
        Void existing preauthorization.

        :param transaction_id:
        :param message:
        :return: status code
        """
        request_data = {
            "message": message
        }

        url = "%s%s%s/void" % (self.api_endpoint, constants.TRANSACTION_STATUS_ENDPOINT, transaction_id)

        username = self.base.get_username()
        password = self.base.get_password(username=username, request_url=url)

        response = requests.put(url, json=request_data, auth=HTTPBasicAuth(username=username, password=password))

        if response.status_code == 404:
            raise TransactionDoesNotExist('Wrong transaction ID!')

        if not self.base.verify_response(response.json()):
            raise SignatureValidationException('Server signature verification has failed')

        response_json = response.json()

        return response_json.get('status')

    def refund_transaction(self, transaction_id, refund_amount):
        """
        Refund the transaction - refund only the amount specified.
        :param transaction_id:
        :return:
        """

        request_data = {
            "amount": self.base.convert_decimal_to_hundreds(refund_amount),
            "currency": self.currency,
        }

        url = f"{self.api_endpoint}{constants.TRANSACTION_STATUS_ENDPOINT}{transaction_id}/refund"

        username = self.base.get_username()
        password = self.base.get_password(username=username, request_url=url)

        response = requests.post(url, json=request_data, auth=HTTPBasicAuth(username=username, password=password))

        if response.status_code == 404:
            raise TransactionDoesNotExist('Wrong transaction ID!')

        if not self.base.verify_response(response.json()):
            raise SignatureValidationException('Server signature verification has failed')

        response_json = response.json()

        return response_json.get('status')



import time
import random
import hashlib
import warnings
import binascii

from decimal import Decimal, ROUND_FLOOR

import rsa


class MBillsBase(object):
    def __init__(self, api_key, shared_secret, mbills_rsa_pub_key, nonce_length=15):
        self.api_key = api_key
        self.shared_secret = shared_secret

        self._pub_key = rsa.PublicKey.load_pkcs1_openssl_pem(mbills_rsa_pub_key) if mbills_rsa_pub_key else None

        assert 8 <= nonce_length <= 15, "Nonce must be of length between 8-15."
        self._nonce_length = nonce_length

    def _get_nonce(self):
        """
        Generate nonce per API specification. A random number of lenght 8-15
        :return: integer of specified length (int)
        """
        return random.randint(10**(self._nonce_length-1), (10**self._nonce_length)-1)

    def get_username(self):
        """
        Construct username as api_key.nonce.timestamp
        :return: username (string)
        """
        return "%s.%d.%d" % (self.api_key, self._get_nonce(), int(time.time()))

    def get_password(self, username, request_url):
        """
        Calculate the password as per the API specification.
        :param username: 
        :param request_url:
        :return: password (string)
        """
        content = "%s%s%s" % (username, self.shared_secret, request_url)
        return hashlib.sha256(content.encode('utf8')).hexdigest()

    def verify_signature(self, signature, nonce, timestamp, signed_id):
        """
        Verify the server response signature. 
        
        :param signature: 
        :param nonce: 
        :param timestamp: 
        :param signed_id: either transactionid, documentid or marketplacemerchantid
        :return: true of false
        """
        message = "%s%s%s%s" % (self.api_key, nonce, timestamp, signed_id)

        if self._pub_key is None:
            warnings.warn("Skipping RSA signature verification. Please pass public key on class initialization to ensure highest level of security!")
            return True

        try:
            rsa.verify(message.encode('utf-8'), signature, self._pub_key)
        except rsa.VerificationError:
            return False

        return True

    def verify_response(self, response_json, signed_id_name='transactionid'):
        """
        Verify the response message.

        :param response_json: 
        :param signed_id_name: 
        :return: 
        """
        auth_json = response_json.get('auth', {})

        nonce = auth_json.get('nonce', '')
        timestamp = auth_json.get('timestamp', '')
        signature = binascii.unhexlify(auth_json.get('signature', ''))

        signed_id = response_json.get(signed_id_name, '')

        return self.verify_signature(signature=signature, nonce=nonce, timestamp=timestamp, signed_id=signed_id)

    def convert_decimal_to_hundreds(self, amount):
        """
        Convert Decimal(10.10) to string "1010"
        :param amount: 
        :return: 
        """
        return str((amount.quantize(Decimal('.01'), rounding=ROUND_FLOOR) * 100).quantize(Decimal('0')))



from distutils.core import setup
setup(
  name='python_mbills',
  packages=['python_mbills'],
  version='0.1.6',
  description='Python implementation of the Hal MBills APIs.',
  author='Boris Savic',
  author_email='boris70@gmail.com',
  url='https://github.com/boris-savic/python-mbills',
  download_url='https://github.com/boris-savic/python-mbills/archive/0.1.6.tar.gz',
  keywords=['mbills', 'halcom', ],
  classifiers=[],
  install_requires=[
        'requests>=2.13.0',
        'rsa>=3.4.0'
    ]
)